package net.sergeych.crypto2

import com.ionspin.kotlin.crypto.secretbox.SecretBox
import kotlinx.serialization.Serializable
import net.sergeych.bintools.toDataSource
import net.sergeych.bipack.BipackDecoder
import net.sergeych.bipack.BipackEncoder
import net.sergeych.crypto2.SymmetricKey.Companion.random
import kotlin.random.Random
import kotlin.random.nextInt
import kotlin.random.nextUBytes

/**
 * Symmetric key implements authenticated encrypting with random nonce and optional fill.
 * Random fill is normally used when cryptanalysis of the message size is a threat.
 *
 * Do not call this constructor directly, use [random] or deserialize it.
 *
 * __Algorithms:__
 *
 * - Encryption: XSalsa20 stream cipher.
 * - Authentication: Poly1305 MAC
 */
@Serializable
class SymmetricKey(
    val keyBytes: UByteArray
): EncryptingKey, DecryptingKey {

    /**
     * @suppress
     * nonce + ciphered data serialization aid
     */
    @Serializable
    data class WithNonce(
        val cipherData: UByteArray,
        val nonce: UByteArray,
    )

    /**
     * @suppress
     * add some random data to the end of the plain message
     */
    @Serializable
    data class WithFill(
        val data: UByteArray,
        val safetyFill: UByteArray? = null
    )

    override fun encrypt(plainData: UByteArray,randomFill: IntRange?): UByteArray {
        val fill = randomFill?.let {
            require(it.start >= 0)
            Random.nextUBytes(Random.nextInt(it))
        }
        val filled = BipackEncoder.encode(WithFill(plainData, fill))
        val nonce = randomNonce()
        val encrypted = SecretBox.easy(filled.toUByteArray(), nonce, keyBytes)
        return BipackEncoder.encode(WithNonce(encrypted, nonce)).toUByteArray()
    }

    override fun decrypt(cipherData: UByteArray): UByteArray {
        val wn: WithNonce = BipackDecoder.Companion.decode(cipherData.toDataSource())
        try {
            return BipackDecoder.Companion.decode<WithFill>(
                SecretBox.openEasy(wn.cipherData, wn.nonce, keyBytes).toDataSource()
            ).data
        }
        catch(_: com.ionspin.kotlin.crypto.secretbox.SecretBoxCorruptedOrTamperedDataExceptionOrInvalidKey) {
            throw DecryptionFailedException()
        }
    }

    companion object {
        /**
         * Create a secure random symmetric key.
         */
        fun random() = SymmetricKey(SecretBox.keygen())
    }

}